﻿# Python 3
# Determine if a PoP1 or PoP2 level has broken room links.
from __future__ import print_function, unicode_literals;
import sys;

def check_links(room_links, start_room):
	neighbor_dirs = ((-1,0), (+1,0), (0,-1), (0,+1)); # left,right,up,down
	room_positions = {};
	room_positions[start_room] = (0,0);
	position_rooms = {};
	position_rooms[(0,0)] = start_room;
	room_queue = [start_room];
	while len(room_queue) > 0:
		room = room_queue.pop(0);
		for (neighbor_room,(dx,dy)) in zip(room_links[room], neighbor_dirs):
			(x,y) = room_positions[room];
			neighbor_pos = (x+dx, y+dy);
			if neighbor_room != 0:
				if neighbor_room not in room_positions:
					room_queue.append(neighbor_room);
				if room_positions.setdefault(neighbor_room, neighbor_pos) != neighbor_pos:
					return False; # Same room at different positions. (2)
			if position_rooms.setdefault(neighbor_pos, neighbor_room) != neighbor_room:
				return False; # Same position for different rooms. (1) Or no link between adjacent rooms. (3)
	return position_rooms;

def print_map(position_rooms):
	if not position_rooms:
		print("Broken room links!");
		return;
	ranges = [[minmax(position[coord] for (position,room) in position_rooms.items() if room) for minmax in (min,max)] for coord in (0,1)];
	((xmin,xmax),(ymin,ymax)) = ranges;
	for y in range(ymin, ymax+1):
		for x in range(xmin, xmax+1):
			room = position_rooms.get((x,y));
			if room:
				print(" %2d" % room, end="");
			else:
				print("   ", end="");
		print();

def check_level(filename):
	with open(filename,"rb") as file:
		data = file.read();

	pop_ver = None;
	if data[0:7] == b"POP_LVL": # *.plv file
		pop_ver = data[7];
		data_len = data[0x0E] | (data[0x0F] << 8);
		data_start = 0x13;
		data = data[data_start : data_start+data_len];
	elif len(data) in (0x900,0x901):
		pop_ver = 1;
	elif len(data) == 12025:
		pop_ver = 2;

	if pop_ver == 1:
		links_start = 0x7A0;
		links_end = 0x800;
		startroom_addr = 0x840;
	elif pop_ver == 2:
		links_start = 0x17C0;
		links_end = 0x1840;
		startroom_addr = 0x1860;
	else:
		print("Can't determine PoP version of level!");
		return 1;

	links_data = data[links_start:links_end];
	start_room = data[startroom_addr];

	n_rooms = len(links_data) // 4;
	room_links = [None]*(n_rooms+1);
	for ix in range(1,n_rooms+1):
		room_links[ix] = list(links_data[(ix-1)*4 : ix*4]);

	print_map(check_links(room_links, start_room));

def main(argv):
	if len(argv) <= 1:
		print("You must pass the filename of a level file in the command line.");
		return 1;
	return check_level(argv[1]);

sys.exit(main(sys.argv));
